/*
 * Copyright (c) 2017 Digital Bazaar, Inc. All rights reserved.
 */
'use strict';

var _createClass = function () { function defineProperties(target, props) { for (var i = 0; i < props.length; i++) { var descriptor = props[i]; descriptor.enumerable = descriptor.enumerable || false; descriptor.configurable = true; if ("value" in descriptor) descriptor.writable = true; Object.defineProperty(target, descriptor.key, descriptor); } } return function (Constructor, protoProps, staticProps) { if (protoProps) defineProperties(Constructor.prototype, protoProps); if (staticProps) defineProperties(Constructor, staticProps); return Constructor; }; }();

function _classCallCheck(instance, Constructor) { if (!(instance instanceof Constructor)) { throw new TypeError("Cannot call a class as a function"); } }

module.exports = function () {
  /**
   * Creates a simple document cache that retains documents for a short
   * period of time.
   *
   * FIXME: Implement simple HTTP caching instead.
   *
   * @param options the options to use:
   *          [size] the maximum size of the cache (default: 50).
   *          [expires] expiration time for each entry in ms (default: 30000).
   */
  function DocumentCache(_ref) {
    var _ref$size = _ref.size,
        size = _ref$size === undefined ? 50 : _ref$size,
        _ref$expires = _ref.expires,
        expires = _ref$expires === undefined ? 30000 : _ref$expires;

    _classCallCheck(this, DocumentCache);

    this.order = [];
    this.cache = {};
    this.size = size;
    this.expires = expires;
  }

  _createClass(DocumentCache, [{
    key: 'get',
    value: function get(url) {
      if (url in this.cache) {
        var entry = this.cache[url];
        if (entry.expires >= Date.now()) {
          return entry.ctx;
        }
        delete this.cache[url];
        this.order.splice(this.order.indexOf(url), 1);
      }
      return null;
    }
  }, {
    key: 'set',
    value: function set(url, ctx) {
      if (this.order.length === this.size) {
        delete this.cache[this.order.shift()];
      }
      this.order.push(url);
      this.cache[url] = { ctx: ctx, expires: Date.now() + this.expires };
    }
  }]);

  return DocumentCache;
}();