/*
 * Copyright (c) 2017 Digital Bazaar, Inc. All rights reserved.
 */
'use strict';

function _asyncToGenerator(fn) { return function () { var gen = fn.apply(this, arguments); return new Promise(function (resolve, reject) { function step(key, arg) { try { var info = gen[key](arg); var value = info.value; } catch (error) { reject(error); return; } if (info.done) { resolve(value); } else { return Promise.resolve(value).then(function (value) { step("next", value); }, function (err) { step("throw", err); }); } } return step("next"); }); }; }

var graphTypes = require('./graphTypes');
var types = require('./types');
var util = require('./util');

// constants

var _require = require('./constants'),
    RDF_LIST = _require.RDF_LIST,
    RDF_FIRST = _require.RDF_FIRST,
    RDF_REST = _require.RDF_REST,
    RDF_NIL = _require.RDF_NIL,
    RDF_TYPE = _require.RDF_TYPE,
    XSD_BOOLEAN = _require.XSD_BOOLEAN,
    XSD_DOUBLE = _require.XSD_DOUBLE,
    XSD_INTEGER = _require.XSD_INTEGER,
    XSD_STRING = _require.XSD_STRING;

var api = {};
module.exports = api;

/**
 * Converts an RDF dataset to JSON-LD.
 *
 * @param dataset the RDF dataset.
 * @param options the RDF serialization options.
 *
 * @return a Promise that resolves to the JSON-LD output.
 */
api.fromRDF = function () {
  var _ref2 = _asyncToGenerator(regeneratorRuntime.mark(function _callee(dataset, _ref) {
    var _ref$useRdfType = _ref.useRdfType,
        useRdfType = _ref$useRdfType === undefined ? false : _ref$useRdfType,
        _ref$useNativeTypes = _ref.useNativeTypes,
        useNativeTypes = _ref$useNativeTypes === undefined ? false : _ref$useNativeTypes;

    var defaultGraph, graphMap, referencedOnce, _iteratorNormalCompletion, _didIteratorError, _iteratorError, _iterator, _step, quad, name, nodeMap, s, p, o, node, objectIsNode, value, object, graphObject, nil, _iteratorNormalCompletion2, _didIteratorError2, _iteratorError2, _iterator2, _step2, usage, property, head, list, listNodes, nodeKeyCount, _iteratorNormalCompletion3, _didIteratorError3, _iteratorError3, _iterator3, _step3, listNode, result, subjects, _iteratorNormalCompletion4, _didIteratorError4, _iteratorError4, _iterator4, _step4, subject, _node, graph, _graphObject, graphSubjects, _iteratorNormalCompletion5, _didIteratorError5, _iteratorError5, _iterator5, _step5, graphSubject, _node2;

    return regeneratorRuntime.wrap(function _callee$(_context) {
      while (1) {
        switch (_context.prev = _context.next) {
          case 0:
            defaultGraph = {};
            graphMap = { '@default': defaultGraph };
            referencedOnce = {};
            _iteratorNormalCompletion = true;
            _didIteratorError = false;
            _iteratorError = undefined;
            _context.prev = 6;
            _iterator = dataset[Symbol.iterator]();

          case 8:
            if (_iteratorNormalCompletion = (_step = _iterator.next()).done) {
              _context.next = 30;
              break;
            }

            quad = _step.value;

            // TODO: change 'name' to 'graph'
            name = quad.graph.termType === 'DefaultGraph' ? '@default' : quad.graph.value;

            if (!(name in graphMap)) {
              graphMap[name] = {};
            }
            if (name !== '@default' && !(name in defaultGraph)) {
              defaultGraph[name] = { '@id': name };
            }

            nodeMap = graphMap[name];

            // get subject, predicate, object

            s = quad.subject.value;
            p = quad.predicate.value;
            o = quad.object;


            if (!(s in nodeMap)) {
              nodeMap[s] = { '@id': s };
            }
            node = nodeMap[s];
            objectIsNode = o.termType.endsWith('Node');

            if (objectIsNode && !(o.value in nodeMap)) {
              nodeMap[o.value] = { '@id': o.value };
            }

            if (!(p === RDF_TYPE && !useRdfType && objectIsNode)) {
              _context.next = 24;
              break;
            }

            util.addValue(node, '@type', o.value, { propertyIsArray: true });
            return _context.abrupt('continue', 27);

          case 24:
            value = _RDFToObject(o, useNativeTypes);

            util.addValue(node, p, value, { propertyIsArray: true });

            // object may be an RDF list/partial list node but we can't know easily
            // until all triples are read
            if (objectIsNode) {
              if (o.value === RDF_NIL) {
                // track rdf:nil uniquely per graph
                object = nodeMap[o.value];

                if (!('usages' in object)) {
                  object.usages = [];
                }
                object.usages.push({
                  node: node,
                  property: p,
                  value: value
                });
              } else if (o.value in referencedOnce) {
                // object referenced more than once
                referencedOnce[o.value] = false;
              } else {
                // keep track of single reference
                referencedOnce[o.value] = {
                  node: node,
                  property: p,
                  value: value
                };
              }
            }

          case 27:
            _iteratorNormalCompletion = true;
            _context.next = 8;
            break;

          case 30:
            _context.next = 36;
            break;

          case 32:
            _context.prev = 32;
            _context.t0 = _context['catch'](6);
            _didIteratorError = true;
            _iteratorError = _context.t0;

          case 36:
            _context.prev = 36;
            _context.prev = 37;

            if (!_iteratorNormalCompletion && _iterator.return) {
              _iterator.return();
            }

          case 39:
            _context.prev = 39;

            if (!_didIteratorError) {
              _context.next = 42;
              break;
            }

            throw _iteratorError;

          case 42:
            return _context.finish(39);

          case 43:
            return _context.finish(36);

          case 44:
            _context.t1 = regeneratorRuntime.keys(graphMap);

          case 45:
            if ((_context.t2 = _context.t1()).done) {
              _context.next = 125;
              break;
            }

            name = _context.t2.value;
            graphObject = graphMap[name];

            // no @lists to be converted, continue

            if (RDF_NIL in graphObject) {
              _context.next = 50;
              break;
            }

            return _context.abrupt('continue', 45);

          case 50:

            // iterate backwards through each RDF list
            nil = graphObject[RDF_NIL];

            if (nil.usages) {
              _context.next = 53;
              break;
            }

            return _context.abrupt('continue', 45);

          case 53:
            _iteratorNormalCompletion2 = true;
            _didIteratorError2 = false;
            _iteratorError2 = undefined;
            _context.prev = 56;
            _iterator2 = nil.usages[Symbol.iterator]();

          case 58:
            if (_iteratorNormalCompletion2 = (_step2 = _iterator2.next()).done) {
              _context.next = 108;
              break;
            }

            usage = _step2.value;
            node = usage.node;
            property = usage.property;
            head = usage.value;
            list = [];
            listNodes = [];

            // ensure node is a well-formed list node; it must:
            // 1. Be referenced only once.
            // 2. Have an array for rdf:first that has 1 item.
            // 3. Have an array for rdf:rest that has 1 item.
            // 4. Have no keys other than: @id, rdf:first, rdf:rest, and,
            //   optionally, @type where the value is rdf:List.

            nodeKeyCount = Object.keys(node).length;

          case 66:
            if (!(property === RDF_REST && types.isObject(referencedOnce[node['@id']]) && types.isArray(node[RDF_FIRST]) && node[RDF_FIRST].length === 1 && types.isArray(node[RDF_REST]) && node[RDF_REST].length === 1 && (nodeKeyCount === 3 || nodeKeyCount === 4 && types.isArray(node['@type']) && node['@type'].length === 1 && node['@type'][0] === RDF_LIST))) {
              _context.next = 78;
              break;
            }

            list.push(node[RDF_FIRST][0]);
            listNodes.push(node['@id']);

            // get next node, moving backwards through list
            usage = referencedOnce[node['@id']];
            node = usage.node;
            property = usage.property;
            head = usage.value;
            nodeKeyCount = Object.keys(node).length;

            // if node is not a blank node, then list head found

            if (graphTypes.isBlankNode(node)) {
              _context.next = 76;
              break;
            }

            return _context.abrupt('break', 78);

          case 76:
            _context.next = 66;
            break;

          case 78:
            if (!(property === RDF_FIRST)) {
              _context.next = 84;
              break;
            }

            if (!(node['@id'] === RDF_NIL)) {
              _context.next = 81;
              break;
            }

            return _context.abrupt('continue', 105);

          case 81:

            // preserve list head
            if (RDF_REST in graphObject[head['@id']]) {
              head = graphObject[head['@id']][RDF_REST][0];
            }

            list.pop();
            listNodes.pop();

          case 84:

            // transform list into @list object
            delete head['@id'];
            head['@list'] = list.reverse();
            _iteratorNormalCompletion3 = true;
            _didIteratorError3 = false;
            _iteratorError3 = undefined;
            _context.prev = 89;
            for (_iterator3 = listNodes[Symbol.iterator](); !(_iteratorNormalCompletion3 = (_step3 = _iterator3.next()).done); _iteratorNormalCompletion3 = true) {
              listNode = _step3.value;

              delete graphObject[listNode];
            }
            _context.next = 97;
            break;

          case 93:
            _context.prev = 93;
            _context.t3 = _context['catch'](89);
            _didIteratorError3 = true;
            _iteratorError3 = _context.t3;

          case 97:
            _context.prev = 97;
            _context.prev = 98;

            if (!_iteratorNormalCompletion3 && _iterator3.return) {
              _iterator3.return();
            }

          case 100:
            _context.prev = 100;

            if (!_didIteratorError3) {
              _context.next = 103;
              break;
            }

            throw _iteratorError3;

          case 103:
            return _context.finish(100);

          case 104:
            return _context.finish(97);

          case 105:
            _iteratorNormalCompletion2 = true;
            _context.next = 58;
            break;

          case 108:
            _context.next = 114;
            break;

          case 110:
            _context.prev = 110;
            _context.t4 = _context['catch'](56);
            _didIteratorError2 = true;
            _iteratorError2 = _context.t4;

          case 114:
            _context.prev = 114;
            _context.prev = 115;

            if (!_iteratorNormalCompletion2 && _iterator2.return) {
              _iterator2.return();
            }

          case 117:
            _context.prev = 117;

            if (!_didIteratorError2) {
              _context.next = 120;
              break;
            }

            throw _iteratorError2;

          case 120:
            return _context.finish(117);

          case 121:
            return _context.finish(114);

          case 122:

            delete nil.usages;
            _context.next = 45;
            break;

          case 125:
            result = [];
            subjects = Object.keys(defaultGraph).sort();
            _iteratorNormalCompletion4 = true;
            _didIteratorError4 = false;
            _iteratorError4 = undefined;
            _context.prev = 130;
            _iterator4 = subjects[Symbol.iterator]();

          case 132:
            if (_iteratorNormalCompletion4 = (_step4 = _iterator4.next()).done) {
              _context.next = 162;
              break;
            }

            subject = _step4.value;
            _node = defaultGraph[subject];

            if (!(subject in graphMap)) {
              _context.next = 158;
              break;
            }

            graph = _node['@graph'] = [];
            _graphObject = graphMap[subject];
            graphSubjects = Object.keys(_graphObject).sort();
            _iteratorNormalCompletion5 = true;
            _didIteratorError5 = false;
            _iteratorError5 = undefined;
            _context.prev = 142;

            for (_iterator5 = graphSubjects[Symbol.iterator](); !(_iteratorNormalCompletion5 = (_step5 = _iterator5.next()).done); _iteratorNormalCompletion5 = true) {
              graphSubject = _step5.value;
              _node2 = _graphObject[graphSubject];
              // only add full subjects to top-level

              if (!graphTypes.isSubjectReference(_node2)) {
                graph.push(_node2);
              }
            }
            _context.next = 150;
            break;

          case 146:
            _context.prev = 146;
            _context.t5 = _context['catch'](142);
            _didIteratorError5 = true;
            _iteratorError5 = _context.t5;

          case 150:
            _context.prev = 150;
            _context.prev = 151;

            if (!_iteratorNormalCompletion5 && _iterator5.return) {
              _iterator5.return();
            }

          case 153:
            _context.prev = 153;

            if (!_didIteratorError5) {
              _context.next = 156;
              break;
            }

            throw _iteratorError5;

          case 156:
            return _context.finish(153);

          case 157:
            return _context.finish(150);

          case 158:
            // only add full subjects to top-level
            if (!graphTypes.isSubjectReference(_node)) {
              result.push(_node);
            }

          case 159:
            _iteratorNormalCompletion4 = true;
            _context.next = 132;
            break;

          case 162:
            _context.next = 168;
            break;

          case 164:
            _context.prev = 164;
            _context.t6 = _context['catch'](130);
            _didIteratorError4 = true;
            _iteratorError4 = _context.t6;

          case 168:
            _context.prev = 168;
            _context.prev = 169;

            if (!_iteratorNormalCompletion4 && _iterator4.return) {
              _iterator4.return();
            }

          case 171:
            _context.prev = 171;

            if (!_didIteratorError4) {
              _context.next = 174;
              break;
            }

            throw _iteratorError4;

          case 174:
            return _context.finish(171);

          case 175:
            return _context.finish(168);

          case 176:
            return _context.abrupt('return', result);

          case 177:
          case 'end':
            return _context.stop();
        }
      }
    }, _callee, undefined, [[6, 32, 36, 44], [37,, 39, 43], [56, 110, 114, 122], [89, 93, 97, 105], [98,, 100, 104], [115,, 117, 121], [130, 164, 168, 176], [142, 146, 150, 158], [151,, 153, 157], [169,, 171, 175]]);
  }));

  return function (_x, _x2) {
    return _ref2.apply(this, arguments);
  };
}();

/**
 * Converts an RDF triple object to a JSON-LD object.
 *
 * @param o the RDF triple object to convert.
 * @param useNativeTypes true to output native types, false not to.
 *
 * @return the JSON-LD object.
 */
function _RDFToObject(o, useNativeTypes) {
  // convert NamedNode/BlankNode object to JSON-LD
  if (o.termType.endsWith('Node')) {
    return { '@id': o.value };
  }

  // convert literal to JSON-LD
  var rval = { '@value': o.value };

  // add language
  if (o.language) {
    rval['@language'] = o.language;
  } else {
    var type = o.datatype.value;
    if (!type) {
      type = XSD_STRING;
    }
    // use native types for certain xsd types
    if (useNativeTypes) {
      if (type === XSD_BOOLEAN) {
        if (rval['@value'] === 'true') {
          rval['@value'] = true;
        } else if (rval['@value'] === 'false') {
          rval['@value'] = false;
        }
      } else if (types.isNumeric(rval['@value'])) {
        if (type === XSD_INTEGER) {
          var i = parseInt(rval['@value'], 10);
          if (i.toFixed(0) === rval['@value']) {
            rval['@value'] = i;
          }
        } else if (type === XSD_DOUBLE) {
          rval['@value'] = parseFloat(rval['@value']);
        }
      }
      // do not add native type
      if (![XSD_BOOLEAN, XSD_INTEGER, XSD_DOUBLE, XSD_STRING].includes(type)) {
        rval['@type'] = type;
      }
    } else if (type !== XSD_STRING) {
      rval['@type'] = type;
    }
  }

  return rval;
}